use crate::gui_handler::gui::displayable::DisplayableFillType;
use crate::prelude::*;
use anyhow::Result;
use assetpath::AssetPath;
use utilities::prelude::*;

use super::gridinfo::GridInfo;
use super::mandatory::Mandatory;

use std::str::from_utf8;
use std::{
    convert::TryFrom,
    sync::{Arc, RwLock, Weak},
};

use super::validator::{
    cow_to_fill_type, cow_to_fill_type_info, cow_to_path, cow_to_str, str_into,
};

pub struct IconInfo {
    pub id: String,

    pub x_slot: Mandatory<usize>,
    pub y_slot: Mandatory<usize>,

    pub x_dim: u32,
    pub y_dim: u32,

    pub icon: Option<AssetPath>,
    pub margin: Option<u32>,

    pub background_type: Option<FillTypeInfo>,
    background_fill_type: DisplayableFillType,

    pub text_ratio: Option<f32>,
    pub text: RwLock<String>,
    pub text_color: Color,

    pub parent: Weak<GridInfo>,
}

impl IconInfo {
    pub fn new<'a>(
        attributes: quick_xml::events::attributes::Attributes<'a>,
        grid: &Arc<GridInfo>,
    ) -> Result<IconInfo> {
        let mut icon_info = IconInfo {
            id: String::new(),

            x_slot: Mandatory::default(),
            y_slot: Mandatory::default(),

            x_dim: 1,
            y_dim: 1,

            icon: None,
            margin: None,

            background_type: None,
            background_fill_type: DisplayableFillType::Expand,

            text_ratio: None,
            text: RwLock::new(String::new()),
            text_color: Color::Black,

            parent: Arc::downgrade(grid),
        };

        for attribute_res in attributes {
            let attribute = attribute_res?;

            match attribute.key.into_inner() {
                b"id" => icon_info.id = cow_to_str(attribute.value),
                b"x_slot" => icon_info.x_slot.set(str_into(attribute.value)?),
                b"y_slot" => icon_info.y_slot.set(str_into(attribute.value)?),
                b"x_size" => icon_info.x_dim = str_into(attribute.value)?,
                b"y_size" => icon_info.y_dim = str_into(attribute.value)?,
                b"icon" => icon_info.icon = Some(cow_to_path(attribute.value)),
                b"margin" => icon_info.margin = Some(str_into(attribute.value)?),
                b"background" => {
                    icon_info.background_type = Some(cow_to_fill_type_info(attribute.value))
                }
                b"fill_type" => {
                    icon_info.background_fill_type = cow_to_fill_type(attribute.value)?;
                }
                b"text_color" => {
                    let text = cow_to_str(attribute.value);
                    icon_info.text_color = Color::try_from(text.as_str())?;
                }
                b"text_ratio" => icon_info.text_ratio = Some(str_into(attribute.value)?),
                _ => {
                    return Err(anyhow::Error::msg(format!(
                        "Unsupported attribute in Icon: {}",
                        from_utf8(attribute.key.into_inner())?
                    )))
                }
            }
        }

        if let Some(background_type) = &mut icon_info.background_type {
            match background_type {
                FillTypeInfo::Image(_) => {}
                FillTypeInfo::Color(_, fill_type) => {
                    *fill_type = icon_info.background_fill_type;
                }
                FillTypeInfo::Element(_, fill_type) => {
                    *fill_type = icon_info.background_fill_type;
                }
            }
        }

        Ok(icon_info)
    }
}
