use anyhow::Result;

use std::ops::Deref;
use std::sync::atomic::{AtomicBool, Ordering::SeqCst};

// used when reading a altered value is important
pub struct Mandatory<T: Send + Sync> {
    value: T,
    modified: AtomicBool,
}

impl<T: Send + Sync> Mandatory<T> {
    pub fn new(value: T) -> Mandatory<T> {
        Mandatory {
            value,
            modified: AtomicBool::new(false),
        }
    }

    pub fn set(&mut self, value: T) {
        self.value = value;
        self.modified.store(true, SeqCst);
    }

    pub fn get_ref(&self) -> Result<&T> {
        if self.modified.load(SeqCst) {
            Ok(&self.value)
        } else {
            Err(anyhow::Error::msg("Mandatory value is not set!"))
        }
    }

    pub fn value_ref_mut(&mut self) -> &mut T {
        &mut self.value
    }

    pub fn value_ref(&self) -> &T {
        &self.value
    }

    pub fn manual_modified(&self) {
        self.modified.store(true, SeqCst);
    }
}

impl<T: Send + Sync> Deref for Mandatory<T> {
    type Target = T;

    fn deref(&self) -> &Self::Target {
        &self.value
    }
}

impl<T: Default + Send + Sync> Mandatory<T> {
    pub fn default() -> Mandatory<T> {
        Mandatory {
            value: Default::default(),
            modified: AtomicBool::new(false),
        }
    }
}

impl<T: Copy + Send + Sync> Mandatory<T> {
    pub fn get(&self) -> Result<T> {
        if self.modified.load(SeqCst) {
            Ok(self.value)
        } else {
            Err(anyhow::Error::msg("Mandatory value is not set!"))
        }
    }
}

impl<T: Send + Sync + std::fmt::Debug> std::fmt::Debug for Mandatory<T> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("Mandatory")
            .field("value", &self.value)
            .field("modified", &self.modified)
            .finish()
    }
}
