mod generator;
mod position_only;

use std::{
    collections::HashMap,
    sync::{Arc, Mutex},
};

use anyhow::Result;
use serde::{Deserialize, Serialize};
use utilities::color::Color;
use vulkan_rs::prelude::*;

use self::generator::Generator;

#[derive(Debug, Clone, Copy, Serialize, Deserialize)]
pub enum ElementBorderThickness {
    Pixel(u32),
    Ratio(f32),
}

impl From<u32> for ElementBorderThickness {
    fn from(value: u32) -> Self {
        Self::Pixel(value)
    }
}

impl From<f32> for ElementBorderThickness {
    fn from(value: f32) -> Self {
        Self::Ratio(value)
    }
}

impl PartialEq for ElementBorderThickness {
    fn eq(&self, other: &ElementBorderThickness) -> bool {
        match (self, other) {
            (ElementBorderThickness::Pixel(m), ElementBorderThickness::Pixel(o)) => m == o,
            (ElementBorderThickness::Ratio(m), ElementBorderThickness::Ratio(o)) => m == o,
            _ => false,
        }
    }
}

/// The serialized string looks like:
///
/// {"background_color":"White","border_color":"Black","border_thickness":{"Pixel":5}}
///
#[derive(Debug, Clone, Copy, PartialEq, Serialize, Deserialize)]
pub struct ElementDescriptor {
    background_color: Color,
    border_color: Color,

    border_thickness: ElementBorderThickness,
}

impl ElementDescriptor {
    pub fn new(
        background_color: impl Into<Color>,
        border_color: impl Into<Color>,
        border_thickness: impl Into<ElementBorderThickness>,
    ) -> Self {
        Self {
            background_color: background_color.into(),
            border_color: border_color.into(),
            border_thickness: border_thickness.into(),
        }
    }
}

#[test]
fn serde_roundtrip() {
    let descriptor = ElementDescriptor::new(
        Color::try_from("#1540e6").unwrap(),
        Color::try_from("#0c1d5f").unwrap(),
        5,
    );

    let s = serde_json::to_string(&descriptor).unwrap();
    let res = serde_json::from_str(&s).unwrap();

    assert_eq!(descriptor, res);
}

#[test]
fn serde_serialization() {
    let s = "{\"background_color\":\"#1540e6\",\"border_color\":\"#0c1d5f\",\"border_thickness\":{\"Pixel\":5}}";

    let res = serde_json::from_str(&s).unwrap();

    let descriptor = ElementDescriptor::new(
        Color::try_from("#1540e6").unwrap(),
        Color::try_from("#0c1d5f").unwrap(),
        5,
    );

    assert_eq!(descriptor, res);
}

#[derive(Debug, Hash, Clone, Copy, PartialEq, Eq)]
struct ElementDefinition {
    width: u32,
    height: u32,

    background_color: Color,
    border_color: Color,
    border_thickness: u32,
}

pub struct ElementCreator {
    generator: Generator,
    elements: HashMap<ElementDefinition, Arc<Image>>,
}

impl ElementCreator {
    pub fn new(device: Arc<Device>, queue: Arc<Mutex<Queue>>) -> Result<Self> {
        Ok(Self {
            generator: Generator::new(device, queue)?,
            elements: HashMap::new(),
        })
    }

    pub fn get(
        &mut self,
        width: u32,
        height: u32,
        descriptor: ElementDescriptor,
    ) -> Result<Arc<Image>> {
        let definition = ElementDefinition {
            width,
            height,

            background_color: descriptor.background_color,
            border_color: descriptor.border_color,

            border_thickness: match descriptor.border_thickness {
                ElementBorderThickness::Pixel(pixel) => pixel,
                ElementBorderThickness::Ratio(ratio) => {
                    (width.min(height) as f32 * ratio).round() as u32
                }
            },
        };

        match self.elements.get(&definition) {
            Some(element) => Ok(element.clone()),
            None => {
                let element = self.generator.generate(definition)?;
                self.elements.insert(definition, element.clone());

                Ok(element)
            }
        }
    }
}
